/*
    Various bitmap colourspace conversion routines:
    - I420toRGB
    - RGBtoI420


*/


#include "../config.h"
#include "../src/drip.h"
extern "C" {
#include "../libvo/yuv2rgb.h"
}

#define SCALEBITS 8
#define ONE_HALF  (1 << (SCALEBITS - 1))
#define FIX(x)          ((int) ((x) * (1L<<SCALEBITS) + 0.5))

/* Lookup tables */
gdouble T1164[256];
gdouble T2018[256];
gdouble T0813[256];
gdouble T1596[256];
gdouble T0391[256];
/* Init done flag */
static gboolean init_done = FALSE;


/* Limit colour to valid range */
inline guint8 normalize_colour(gdouble colour) {
    static guint colour_norm;

    if (colour<0) {
        colour_norm = 0;
    } else if (colour>255) {
        colour_norm = 255;
    } else {
        colour_norm = (guint8)(colour);
    }
    return colour_norm;
}
       

/* TODO no chroma interpolating is done atm */
void I420toRGB24(guint8 *dst, guint8 *src, gint width, gint height) {
    register gint offset = 0;
    register gint offset4 = 0;
    register gint offset3 = 0;
    static gdouble offset4_f;
    register guint8 Y,CR,CB;
    register gdouble Bf,Gf,Rf;
    static guint8 *src0,*src1,*src2;
 
    src0 = src;
    src1 = (guint8*)((glong)src0+width*height);
    src2 = (guint8*)((glong)src1+width*height/4);

    offset = 0;
    offset3 = 0;
    /* Loop */
    for (register gint y=0;y<height;y++) {
        offset4_f = (y>>1) * (width>>1);
        for (register gint x=0;x<width;x++) {
             offset4 = (gint)offset4_f;

             Y  = src0[offset];
             CB = src1[offset4];
             CR = src2[offset4];
 
             Bf = T1164[Y] + T2018[CB];
             Gf = T1164[Y] - T0813[CR] - T0391[CB];
             Rf = T1164[Y] + T1596[CR];
 
             dst[offset3++] = normalize_colour(Rf); //(B) + (G<<8) + (R<<16); // RGB 24bits
             dst[offset3++] = normalize_colour(Gf);
             dst[offset3++] = normalize_colour(Bf);
 
             offset++;
             offset4_f = offset4_f+0.5;
        }
    }

    /* Clean & Exit */
    return;
}   

void I420toRGB24_init(void) {
    /* Calculate lookup tables */
    for (gdouble i=-16;i<240;i++) {
        T1164[(gint)(i+16)] = i*1.164;
    }
    for (gdouble i=-128;i<128;i++) {
        T2018[(gint)(i+128)] = i*2.018;
        T0813[(gint)(i+128)] = i*0.813;
        T1596[(gint)(i+128)] = i*1.596;
        T0391[(gint)(i+128)] = i*0.391;
    }
    init_done = TRUE;
    return;
}


/* TODO no chroma interpolating is done atm */
void RGB24toI420(guint8 *dst, guint8 *src, gint width, gint height) {
/*
static void rgb24_to_yuv420p(UINT8 *lum, UINT8 *cb, UINT8 *cr,
                              UINT8 *src, int width, int height)
{ */
    gint wrap, wrap3, x, y;
    gint r, g, b, r1, g1, b1;
    guint8 *p;
    guint8 *lum, *cb, *cr;
    
    lum = dst;
    cb = (guint8*)((glong)dst + width*height);
    cr = (guint8*)((glong)cb + width*height/4);
  
    wrap = width;
    wrap3 = width * 3;
    p = src;
    for(y=0;y<height;y+=2) {
        for(x=0;x<width;x+=2) {
            r = p[0];
            g = p[1];
            b = p[2];
            r1 = r;
            g1 = g;
            b1 = b;
            lum[0] = (FIX(0.29900) * r + FIX(0.58700) * g +
                      FIX(0.11400) * b + ONE_HALF) >> SCALEBITS;
            r = p[3];
            g = p[4];
            b = p[5];
            r1 += r;
            g1 += g;
            b1 += b;
            lum[1] = (FIX(0.29900) * r + FIX(0.58700) * g +
                      FIX(0.11400) * b + ONE_HALF) >> SCALEBITS;
            p += wrap3;
            lum += wrap;

            r = p[0];
            g = p[1];
            b = p[2];
            r1 += r;
            g1 += g;
            b1 += b;
            lum[0] = (FIX(0.29900) * r + FIX(0.58700) * g +
                      FIX(0.11400) * b + ONE_HALF) >> SCALEBITS;
            r = p[3];
            g = p[4];
            b = p[5];
            r1 += r;
            g1 += g;
            b1 += b;
            lum[1] = (FIX(0.29900) * r + FIX(0.58700) * g +
                      FIX(0.11400) * b + ONE_HALF) >> SCALEBITS;

            cb[0] = ((- FIX(0.16874) * r1 - FIX(0.33126) * g1 +
                      FIX(0.50000) * b1 + 4 * ONE_HALF - 1) >> (SCALEBITS + 2)) + 128;
            cr[0] = ((FIX(0.50000) * r1 - FIX(0.41869) * g1 -
                     FIX(0.08131) * b1 + 4 * ONE_HALF - 1) >> (SCALEBITS + 2)) + 128;

            cb++;
            cr++;
            p += -wrap3 + 2 * 3;
            lum += -wrap + 2;
        }
        p += wrap3;
        lum += wrap;
    }
    return;
}  

void RGB24toI420_init(void) {
    /* Calculate lookup tables */
    return;
}

